%SUPCK algorithm for binding pocket comaprison (core of the PARIS server www.cbio.ensmp.fr/paris)
%Copyright 2009 Zaslavskiy Mikhail (mikhail.zaslavskiy@mines-paristech.fr)
%This script may be called as a stand-alone application, or as a sub-function
%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%[val,val_f]=calc_scores(dist_type, exp_sigma,fplist,pathout,fsubmatrix,fnumlabs,tweights,maxradius,chpart_kernel_sigma,blcomparison,selcolumns,radfreq,algo,emulate_sitesbase,opttol,maxpermutation)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%Input:
%--dist_type may be '3D_kernel', '3D_kernel_nd' (default),'3D_distance','3D_norm_distance'
%-------------- '3D_kernel' maximize the sup_CK kernel  over y_t and R, return the value of kernel
%-------------- '3D_kernel_nd' maximize the normalized sup_CK kernel:
%sup_CK(P1,P2)/sqrt(sup_CK(P1,P1)sup_CK(P1,P2))
%-------------- '3D_distance', '3D_norm_distance'  - operate with distance value instead of kernel
%-------------- Recomended choice: '3D_kernel_nd'
%--exp_sigma : \sigma in the sup_CK kernel
%--fplist - a txt file with names of files containings binding pockets
%--pathout - output directory
%--fsubmatrix and fnumlabs - define similarity measure between pocket atoms, 
%		-- if fnumlabs='ChPart', then a Gaussian kernel on atom partial charges with sigma equals to 
%				chpart_kernel_sigma is used
%		-- if fnumlabs='No', then we use a kernel matrix stored in fsubmatrix, it may be an Indentity matrix (it corr%				esponds to the dirac kernel between atoms), a matrix with ones everywere (all atoms are equivalent			     pure geometrical approach), or any user-defined kernel matrix. Each row (column) of the kernels ma				  trix corresponds to  a particualr atom type (label).
%--tweights -- may be used to weight atom important depending on the atom radius, e^(-tweights*r), tweights=0 -- no weighting
%		(default value)
%--maxradius -- pocket extraction parameter, we consider only pocket atoms within maxradius around ligand
%--chpart_kernel_sigma --see fnumlabs
%--blcomparison - 0 (pocket comparison), 1 (ligand comparioson)
%--selcolumns - defines subset of pockets which are processed. 0 - all-vs-all matrix is evaluated, i - i-th kernel matrix row is evaluated, (-2) - kernel matrix diagonal is evaluated 
%--radfreq -- not important
%--algo: optimization algorithm, 'QN' - quasi-newton, 'FM' - matlab fminsearch
%--emulate_sitesbase: 1 - count the number of overlapping  atoms
%--opttol - stop criterion in the quasi-newton algorithm, =-1: very fast
%           estimation based only on the initial guesses (principal axes)
%--maxpermutation -- defines how many intitial points will be tested: 1 (8 initial combinations, 2^3 flips), 2 (16), 3 (24)
%--bonline_load -- 1: read pockets on-line 
%initial combinations, 2^3 flips), 2 (16), 3 (24)
%Output:
%--val is an upper triangular kernel matrix
%--val_f --- debug information
%--in  directory 'pathout', script stocks

function [val,val_f]=calc_scores(dist_type, exp_sigma,fplist,pathout,fsubmatrix,fnumlabs,tweights,maxradius,chpart_kernel_sigma,blcomparison,selcolumns,radfreq,algo,emulate_sitesbase,opttol,maxpermutation,bonline_load)
path='./';
%fnumlabs='./Data/charges_partielles.txt'; %pocket numerical labels association (real numbers)
%fnumlabs='ChPart'; %pocket numerical labels association (real numbers)
%emulate_sitesbase: an algorithm for emulation of the sitesbase algorithm (Poisson index)
i=0;
if nargin<2
    exp_sigma=30;
end
if nargin<3
    fplist='pocket_list.txt';
end
if nargin<4
    pathout='./';
end
if nargin<5
    fsubmatrix='./Data/chemical_kernel.txt';
end
if nargin<6
    fnumlabs='No';
end
if nargin<7
    tweights=0;
else
    if ~isnumeric(tweights)
   	 tweights=str2num(tweights);
    end
end
if nargin<8
    maxradius=-1;
else
    if ~isnumeric(maxradius)
       	 maxradius=str2num(maxradius);
    end
end
if nargin<9
    chpart_kernel_sigma=-1;%linear kernel
else
    if ~isnumeric(chpart_kernel_sigma)
       	 chpart_kernel_sigma=str2num(chpart_kernel_sigma);
    end
end
if nargin<10
    blcomparison=0;%pocket comparison
else
    if ~isnumeric(blcomparison)
       	 blcomparison=str2num(blcomparison);
    end
end
if nargin<11
    selcolumns=-1;%all versus all
else
    if ~isnumeric(selcolumns)
       	 selcolumns=str2num(selcolumns);
    end
end

if nargin<12
    radfreq=3.5;%all versus all
else
    if ~isnumeric(radfreq)
       	 radfreq=str2num(radfreq);
    end
end
if nargin<13
    algo='QN';%all versus all
end
if nargin<14
    emulate_sitesbase=0; %all versus all
else
    if ~isnumeric(emulate_sitesbase)
       	 emulate_sitesbase=str2num(emulate_sitesbase);
    end
end
if nargin<15
    opttol=1e-4; 
else
    if ~isnumeric(opttol)
       	 opttol=str2num(opttol);
    end
end
if nargin<16
    maxpermutation=3; 
else
    if ~isnumeric(maxpermutation)
       	 maxpermutation=str2num(maxpermutation);
    end
end
if nargin<17
    bonline_load=0; 
else
    if ~isnumeric(bonline_load)
       	 bonline_load=str2num(bonline_load);
    end
end
flist=textread(fplist,'%s');
llist=[];LigLst=[];
%loading
for i=1:length(flist) 
    flist{i}
    disp(i);
    P{i}=load_pocket([path flist{i}],fnumlabs,tweights,maxradius,blcomparison);
    LigLst=[LigLst;P{i}.lig];
    if bonline_load
        P{i}=0;
    end
end
f=fopen([pathout '/' 'LigLst.txt'],'w');fprintf(f,'%c%c%c\n',LigLst');fclose(f);
%Ligand enumeration
clust(1)=1;
for i=2:size(LigLst)
newl=-1;
for j=1:i-1
        if strcmp(LigLst(i,:),LigLst(j,:))
                newl=clust(j);
                break;  
        end
end
if newl==-1
        clust(i)=max(clust)+1;
else
        clust(i)=newl;
end
end
Lclass=clust; ULclass=unique(clust);

norms_vector=false;
if (selcolumns==-2) norms_vector=true; end;%output pocket norms
if (selcolumns<1) selcolumns=1:length(P); end;

%construct substitiution matrix for vertices
Llabel=[];
if (blcomparison)
    for i=1:length(P)
        Llabel=[Llabel P{i}.llabels];
    end
    ULlabel=unique(Llabel);
    for i=1:length(P)
       [t P{i}.labels_num]=ismember(P{i}.llabels,ULlabel);
    end
    SubMatrix=eye(length(ULlabel))
end
    
if strcmp(fnumlabs,'SimA') && (~blcomparison)
    for i=1:length(P)
        Llabel=[Llabel P{i}.labels_num'];
    end
    ULlabel=unique(Llabel);
    for i=1:length(P)
       [t P{i}.labels_num]=ismember(P{i}.labels_num,ULlabel);
    end
    SubMatrix=eye(length(ULlabel))
end

if (strcmp(fsubmatrix,'Freq') || strcmp(fnumlabs,'ChComp'))
%list of all labels
rad_inter=radfreq;
Ulabels(1).name='beg';Llabels=[];
for i=1:length(P)
	Llabels=[Llabels P{i}.labels];
	P{i}.labels_num=[];
	for j=1:length(P{i}.labels)
		P{i}.labels(j).name=P{i}.labels(j).name(1:end);
		k=1;bfound=false;
		while (k<=length(Ulabels)) & (~bfound)
			if (strcmp(Ulabels(k).name,P{i}.labels(j).name)==1)
				P{i}.labels_num(j)=k;
				bfound=true;
			end
			k=k+1;
		end
		if (~bfound)
			Ulabels(end+1)=P{i}.labels(j);
		        P{i}.labels_num(j)=k;
		end
	end
disp([num2str(unique(P{i}.labels_num))]);
end

SubMatrix=zeros(length(Ulabels),length(Ulabels));
if strcmp(fnumlabs,'ChComp')
    SubMatrix=eye(size(SubMatrix))
else
for l=1:length(ULclass)
    il=find(Lclass==ULclass(l));
    for ipdb=1:length(P{il(1)}.lpdb)
        ip=[];
        for i=1:length(P)
            if Lclass(i)==ULclass(l)
                for ill=1:length(P{i}.lpdb)
                    if strcmp(P{il(1)}.lpdb(ipdb).pdb1(14:16),P{i}.lpdb(ill).pdb1(14:16))
                        K=P{i}.coords*(P{i}.lcoords(ipdb,:)');
                        Dp=sum(P{i}.coords.^2');
                        Dl=sum(P{i}.lcoords(ipdb,:).^2');
                        D=repmat(Dp',1,length(Dl))+repmat(Dl,length(Dp),1)-2*K;

                        D=(sqrt(D)<rad_inter);
                        ipt=find(D);
                        ip=[ip P{i}.labels_num(ipt)];
                    end
                end
            end
        end
        ipu=unique(ip);ipm=[];
        for t1=1:length(ipu)
            for t2=1:length(ipu)
                ipm(t1,t2)=sum(ip==ipu(t1))*sum(ip==ipu(t2));
            end
        end
        SubMatrix(ipu,ipu)=SubMatrix(ipu,ipu)+ipm;
    end
end
end
%normalization
for i=1:size(SubMatrix,1)
    if SubMatrix(i,i)>0
        t=SubMatrix(i,i);
        SubMatrix(i,:)=SubMatrix(i,:)/sqrt(t);
        SubMatrix(:,i)=SubMatrix(:,i)/sqrt(t);
    end
end
end %frequency kernel
if 0%temporal stuff
M=[];
for i=1:length(P)
%[V1 D1]=eigs(cov(P{i}.lcoords.*repmat(P{i}.weights',1,3)));
[V1 D1]=eigs(cov(P{i}.coords),3);
P_desc(i,:)=diag(D1)';
disp([LigLst(i,:) ' ' flist{i} ' ' num2str([diag(D1)' prod(diag(D1))])]);
 for j=1:0;length(P)
% M(i,j)=abs(size(P{i}.coords,1)-size(P{j}.coords,1));
%[V2 D2]=eigs(cov(P{j}.lcoords.*repmat(P{j}.weights',1,3)));
[V2 D2]=eigs(cov(P{j}.coords),3);disp(j);
%prod(D1)
%prod(D2)
%M(i,j)=abs(prod(diag(D1))-prod(diag(D2)));
%M(i,j)=sum(abs(diag(D1)-diag(D2)));
%M(i,j)=length(setxor(P{i}.labels_num,P{j}.labels_num));
 end
 end
 %save('PTB_ell.txt','M','-ASCII');
 P_desc=[P_desc Lclass'];
 save('PTB_data_compl_rev.txt','P_desc','-ASCII');
return
end
%scores
Nl=length(flist);
options.num_step_rot=5; 
options.tol=10;
options.exp_sigma=str2num(exp_sigma);
options.min.max_it=10;
options.min.tol=1e-8;
%options.optparam.tol_df=opttol_df;
options.optparam.tol=opttol;
options.optparam.kmax=1000;
options.optparam.display=1;
options.rand_init=0;
options.algo=algo;
options.sitesbase_emulation=emulate_sitesbase;
options.maxpermutation=maxpermutation;

if ~strcmp(fnumlabs,'ChPart')
    if strcmp(fsubmatrix,'Freq') || strcmp(fnumlabs,'SimA')
	    options.Kp=SubMatrix;
    else
	    options.Kp=load(fsubmatrix);
    end
else
    all_values=[];
    %we load the kernel from pocket labels
    for i=1:length(P)
        all_values=[all_values P{i}.labels_num];
    end
    all_values_unique=unique(all_values);
    for i=1:length(P)
        [t P{i}.labels_num]=ismember(P{i}.labels_num,all_values_unique);
    end
    if (chpart_kernel_sigma<=0)
        options.Kp=all_values_unique'*all_values_unique;
    else
        options.Kp=(repmat(all_values_unique',1,length(all_values_unique))-repmat(all_values_unique',1,length(all_values_unique))').^2;
        options.Kp=exp(-options.Kp./(2*chpart_kernel_sigma^2));
    end
        
end
options.kernel_output=0;
if strcmp(dist_type,'3D_kernel')
	dist_function=@kernel_3D_exp_prod;
	options.fmeasure_grad=@kernel_3D_exp_prod_grad;
end
if strcmp(dist_type,'3D_kernel_nd') %optimization on the basis of normalized distance and then output the value of the kernel function
	dist_function=@distance_norm_3D_exp_prod;
	options.fmeasure_grad=@distance_norm_3D_exp_prod_grad;
    options.kernel_output=1;
end

if strcmp(dist_type,'3D_distance')
	dist_function=@distance_3D_exp_prod;
	options.fmeasure_grad=@distance_3D_exp_prod_grad;
end

if strcmp(dist_type,'3D_norm_distance')
	dist_function=@distance_norm_3D_exp_prod;
	options.fmeasure_grad=@distance_norm_3D_exp_prod_grad;
end

if strcmp(dist_type,'3D_kernel_lap')
	dist_function=@kernel_3D_lap_prod;
	options.fmeasure_grad=@kernel_3D_lap_prod_grad;
end

if strcmp(dist_type,'3D_kernel_max')
	dist_function=@kernel_3D_exp_prod_max;
    options.fmeasure_grad=@kernel_3D_lap_prod_grad;
end
%addpath('~/graphmatching/src/');
for ii=1:length(selcolumns)
    i=selcolumns(ii);
    if bonline_load
        P=[];
        P{i}=load_pocket([path flist{i}],fnumlabs,tweights,maxradius,blcomparison);
    end
    if norms_vector
        vall(ii)=kernel_3D_exp_prod(P{i},P{i},options);
        save([pathout '/' dist_type '_' num2str(options.exp_sigma) '_norms'],'vall','-ASCII');
    	length_P=i;
        val(ii)=vall(ii);
        continue;
    else 
	length_P=length(flist);
    end
    for j=i:length_P   
        if bonline_load
            Pt=load_pocket([path flist{j}],fnumlabs,tweights,maxradius,blcomparison);
        else
            Pt=P{j};
        end
        if options.rand_init
            %rotation matrix
            xi=2*pi*rand(1);
            psi=2*pi*rand(1);
            phi=2*pi*rand(1);

            Rf=eye(3,3);R=zeros(3,3);
            R(1:2,1:2)=[cos(phi) sin(phi);-sin(phi) cos(phi)];R(3,3)=1; Qphi=R;Rf=Rf*R;
            R(:)=0;R(2:3,2:3)=[cos(psi) sin(psi);-sin(psi) cos(psi)];R(1,1)=1;Qpsi=R;Rf=Rf*R;
            R(:)=0;R([1 3],[1 3])=[cos(xi) sin(xi);-sin(xi) cos(xi)];R(2,2)=1;Qxi=R;Rf=Rf*R;
            Pt.coords=(Rf*(P{j}.coords'))';
        end
        options.norm_factor=sqrt(kernel_3D_exp_prod(P{i},P{i},options)*kernel_3D_exp_prod(Pt,Pt,options));
        options.add_factor=-kernel_3D_exp_prod(P{i},P{i},options)-kernel_3D_exp_prod(Pt,Pt,options);
        if strcmp(dist_type,'graphm')
            [val(1,j) val_f(1,j)]=pocket_dist_graphm(P{i},Pt,options);
       	else
	    if strcmp(dist_type,'kpca')
            	[val(1,j) val_f(1,j)]=pocket_dist_kpca(P{i},Pt,options);	
	    else
	        %[val(1,j) val_f(1,j)]=pocket_dist_pcom(P{i},Pt,dist_function,options);
             [val(ii,j) val_f]=pocket_dist_pcom(P{i},Pt,dist_function,options);
	    end;
        end
        if bonline_load
            %val_f=[];%save memory if we don't use it
        end
        %[val(i,j) val_f(i,j)]=pocket_dist(P{i},Pt,dist_function,options)
		%disp(selcolumns);   
        if norms_vector
            vall(ii)=val(ii,j);
	        save([pathout '/' dist_type '_' num2str(options.exp_sigma) '_norms'],'vall','-ASCII');
        else	
            %vall=val(ii,:);
            vall=val(ii,:);
	    save([pathout '/' dist_type '_' num2str(options.exp_sigma) '_' num2str(selcolumns(ii))],'vall','-ASCII');
        end
        [t flist1]=fileparts(flist{i});
        [t flist2]=fileparts(flist{j});
        disp([j length_P]);
        prnt=0;
 	if prnt
        print_pocket(P{i},[pathout '/' flist1],'w+');
        print_pocket(val_f(ii,j).P2_rot,[pathout '/' flist1 '_' flist2],'w+');
        Rv=[val_f(ii,j).R;Pt.center_coord;val_f(ii,j).v];
        save([pathout '/' flist1 '_' flist2 '_Rv'],'Rv','-ASCII');
	end
    end
    if bonline_load
        P{i}=0;
    end
   %save('dump_all.mat');
end
if ~isempty(strfind(dist_type,'3D_kernel'))
	val=-val;
end
end
%**************************************************************************
%**************************************************************************
%electrostatic poteintial in point x created by the pocket P
function val=esp(P,x)
e0=8.854e-12*1e+19*10^-10;N=size(P,1);
r=sqrt(sum((repmat(x,N,1)-P(:,3:end))'.^2))';
val=sum(P(:,2)./r)/(4*pi*e0);
end

%pocket distance function based on esp, fmeasure - distance measure between two different pocket given fixed rotation angles
function [val,val_f]=pocket_dist(P1,P2,fmeasure,options)
val_rot=[];num_step=options.num_step_rot;x_rot=[];
for phi=-pi:2*pi/(num_step):0.99*pi
    for psi=-pi:2*pi/num_step:0.99*pi
        for xi=-pi:2*pi/num_step:0.99*pi
            %val_rot(end+1)=pocket_dist_rot(P1,P2,fmeasure,phi,psi,xi,options);
            [x exit_param]=minimize_limited_quasi_newton([phi psi xi]',@pocket_dist_rot_full,options.optparam,P1,P2,fmeasure,options);
            val_rot(end+1)=pocket_dist_rot(P1,P2,fmeasure,x(1),x(2),x(3),options);
            x_rot(end+1,:)=x';
        end
    end
end
val=min(val_rot);
val_f.val_rot=val_rot;
val_f.x_rot=x_rot;
end
%pocket distance based on guessing about main axes, try to match principal
%components
function [val,val_f]=pocket_dist_pcom(P1,P2,fmeasure,options)
[V1 D1]=eigs(cov(P1.coords));
[V2 D2]=eigs(cov(P2.coords));
P1.coords=P1.coords*V1;
val_rot=[];x_rot=[];P2_rot=[];fx_descent=[];
Pt=P2;
%try to permute 1-2 and 2-3
permaxes=[1 2 3; 2 1 3; 1 3 2];
for ip=1:min(options.maxpermutation,size(permaxes,1))
%now we have to test all possible -1 +1 combinations
signs=[ 0 0 0;1 0 0;0 1 0;1 1 0; 0 0 1; 1 0 1; 0 1 1; 1 1 1];
for i=1:8
    Pt.coords=P2.coords*V2;Pt.lcoords=P2.lcoords*V2;
    Pt.coords=Pt.coords(:,permaxes(ip,:));Pt.lcoords=Pt.lcoords(:,permaxes(ip,:));
    Pt.coords(:,signs(i,:)==1)=-Pt.coords(:,signs(i,:)==1);Pt.lcoords(:,signs(i,:)==1)=-Pt.lcoords(:,signs(i,:)==1);
    if strcmp(options.algo,'QN')
       %[x exit_param]=minimize_limited_quasi_newton([0 0 0]',@pocket_dist_rot_full,options.optparam,P1,Pt,fmeasure,options);
      %[val_rot(end+1) P2_rot(end+1).P]=pocket_dist_rot(P1,Pt,fmeasure,x(1),x(2),x(3),options);
      if options.optparam.tol>0
        [x exit_param]=minimize_limited_quasi_newton([0 0 0 0 0 0]',@pocket_dist_rot_shift_full,options.optparam,P1,Pt,fmeasure,options);
      else
        x=[0 0 0 0 0 0]';
        exit_param.fx=-1;
      end
       % [x exit_param]=minimize_gradient([0 0 0 0 0 0]',@pocket_dist_rot_shift_full,options.optparam,P1,Pt,fmeasure,options);
        if options.sitesbase_emulation
            [val_rot(end+1) P2_rot(end+1).P ]=pocket_dist_rot_shift_sb_emulate(P1,Pt,fmeasure,x(1),x(2),x(3),x(4),x(5),x(6),options);
        else
            if (options.kernel_output==1)
                s=length(P2_rot);
                [val_rot(end+1) P2_rot(s+1).P P2_rot(s+1).R P2_rot(s+1).v]=pocket_dist_rot_shift(P1,Pt,@kernel_3D_exp_prod,x(1),x(2),x(3),x(4),x(5),x(6),options);
            else
                [val_rot(end+1) P2_rot(end+1).P]=pocket_dist_rot_shift(P1,Pt,fmeasure,x(1),x(2),x(3),x(4),x(5),x(6),options);
            end
        end
    end
    %x=[0;0;0];
    if strcmp(options.algo,'FM')
        global P1;global P2;P2=Pt;global options;global fmeasure;
        [x,fval] = fminsearch(@pocket_dist_rot_glob,[0 0 0]);
        [val_rot(end+1) P2_rot(end+1).P]=pocket_dist_rot_glob(x);
    end
    P2_rot(end).V2=V2;P2_rot(end).permaxes=permaxes(ip,:);P2_rot(end).signs=signs(i,:);
    x_rot(end+1,:)=x';
    fx_descent{end+1}=exit_param.fx;
%    disp(val_rot);
end
end
%disp(val_rot); t=t1;
[val i_min]=min(val_rot);
val_f.val_rot=val_rot;
val_f.x_rot=x_rot;
val_f.P2_rot=P2_rot(i_min).P;
%val_f.R=[P2_rot(i_min).R;P2_rot(i_min).V2;P2_rot(i_min).permaxes;P2_rot(i_min).signs;V1;];
%val_f.v=P2_rot(i_min).v;
%return to the orignal axes of P1
val_f.P2_rot.coords=val_f.P2_rot.coords*V1';
val_f.P2_rot.lcoords=val_f.P2_rot.lcoords*V1';
val_f.P2_rot
val_f.fx_descent=fx_descent;
end%functoin
%kpca dist
function [val,val_f]=pocket_dist_kpca(P1,P2,options)
N=size(P1.coords,1);
ind=repmat(1:N,N,1);ind=ind(:);
K3D=reshape(exp(-sum((repmat(P1.coords,N,1)-P1.coords(ind,:)).^2,2)/(2*options.exp_sigma^2))',N,N).*options.Kp(P1.labels_num,P1.labels_num).*(P1.weights'*P1.weights);
[V D1]=eigs(K3D);
D1=diag(D1);
N=size(P2.coords,1);
ind=repmat(1:N,N,1);ind=ind(:);
K3D=reshape(exp(-sum((repmat(P2.coords,N,1)-P2.coords(ind,:)).^2,2)/(2*options.exp_sigma^2))',N,N).*options.Kp(P2.labels_num,P2.labels_num).*(P2.weights'*P2.weights);
[V D2]=eigs(K3D);
D2=diag(D2);
N=min(length(D1),length(D2));
val=norm(D1(1:N)-D2(1:N),'fro');
val_f=val;
end%functoin

%pocket distance based on guessing about ligand main axes
function [val,val_f]=pocket_dist_pcom_lig(P1,P2,fmeasure,options)
[V1 D1]=eigs(cov(P1.lcoords));
[V2 D2]=eigs(cov(P2.lcoords));
P1.coords=P1.coords*V1;
val_rot=[];x_rot=[];
Pt=P2;
%try to permute 1-2 and 2-3
permaxes=[1 2 3; 2 1 3; 1 3 2];
for ip=1:size(permaxes,2)
%now we have to test all possible -1 +1 combinations
signs=[ 0 0 0;1 0 0;0 1 0;1 1 0; 0 0 1; 1 0 1; 0 1 1; 1 1 1];
for i=1:8
    Pt.coords=P2.coords*V2;
    Pt.coords=Pt.coords(:,permaxes(ip,:));
    Pt.coords(:,signs(i,:)==1)=-Pt.coords(:,signs(i,:)==1);
    [x exit_param]=minimize_limited_quasi_newton([0 0 0]',@pocket_dist_rot_full,options.optparam,P1,Pt,fmeasure,options);
 
    val_rot(end+1)=pocket_dist_rot(P1,Pt,fmeasure,x(1),x(2),x(3),options);
    x_rot(end+1,:)=x';
end
end
val=min(val_rot);
val_f.val_rot=val_rot;
val_f.x_rot=x_rot;
end%functoin

%this function calculate the difference between  two esp under given
%rotation of the second pocket: input just two pockets
function [val,grad]= pocket_dist_rot_full(P1,P2,fmeasure,options,x)
val= pocket_dist_rot(P1,P2,fmeasure,x(1),x(2),x(3),options);
grad= pocket_dist_rot_grad(P1,P2,fmeasure,x(1),x(2),x(3),options);
end

function [val,grad]= pocket_dist_rot_shift_full(P1,P2,fmeasure,options,x)
val= pocket_dist_rot_shift(P1,P2,fmeasure,x(1),x(2),x(3),x(4),x(5),x(6),options);
grad= pocket_dist_rot_shift_grad(P1,P2,fmeasure,x(1),x(2),x(3),x(4),x(5),x(6),options);
end


function [val,P2]= pocket_dist_rot(P1,P2,fmeasure,phi,psi,xi,options)
%rotation matrix
Rf=eye(3,3);R=zeros(3,3);
R(1:2,1:2)=[cos(phi) sin(phi);-sin(phi) cos(phi)];R(3,3)=1; Qphi=R;Rf=Rf*R;
R(:)=0;R(2:3,2:3)=[cos(psi) sin(psi);-sin(psi) cos(psi)];R(1,1)=1;Qpsi=R;Rf=Rf*R;
R(:)=0;R([1 3],[1 3])=[cos(xi) sin(xi);-sin(xi) cos(xi)];R(2,2)=1;Qxi=R;Rf=Rf*R;

P2.coords=(Rf*(P2.coords'))';P2.lcoords=(Rf*(P2.lcoords'))';
val=fmeasure(P1,P2,options);
end

%upgrated version
function [val,P2,Rf,v]= pocket_dist_rot_shift(P1,P2,fmeasure,phi,psi,xi,x,y,z,options)
%rotation matrix
Rf=eye(3,3);R=zeros(3,3);
R(1:2,1:2)=[cos(phi) sin(phi);-sin(phi) cos(phi)];R(3,3)=1; Qphi=R;Rf=Rf*R;
R(:)=0;R(2:3,2:3)=[cos(psi) sin(psi);-sin(psi) cos(psi)];R(1,1)=1;Qpsi=R;Rf=Rf*R;
R(:)=0;R([1 3],[1 3])=[cos(xi) sin(xi);-sin(xi) cos(xi)];R(2,2)=1;Qxi=R;Rf=Rf*R;

P2.coords=(Rf*(P2.coords'))';P2.lcoords=(Rf*(P2.lcoords'))';
%shift
P2.coords=P2.coords+repmat([x y z],size(P2.coords,1),1);P2.lcoords=P2.lcoords+repmat([x y z],size(P2.lcoords,1),1);
v=[x y z];
val=fmeasure(P1,P2,options);%kernel_3D_exp_prod(P1,P2,options);%
end

%here we threeshold distances to emulate the behavior of the sites_base
%approach
function [val,P2]= pocket_dist_rot_shift_sb_emulate(P1,P2,fmeasure,phi,psi,xi,x,y,z,options)
%rotation matrix
Rf=eye(3,3);R=zeros(3,3);
R(1:2,1:2)=[cos(phi) sin(phi);-sin(phi) cos(phi)];R(3,3)=1; Qphi=R;Rf=Rf*R;
R(:)=0;R(2:3,2:3)=[cos(psi) sin(psi);-sin(psi) cos(psi)];R(1,1)=1;Qpsi=R;Rf=Rf*R;
R(:)=0;R([1 3],[1 3])=[cos(xi) sin(xi);-sin(xi) cos(xi)];R(2,2)=1;Qxi=R;Rf=Rf*R;

P2.coords=(Rf*(P2.coords'))';P2.lcoords=(Rf*(P2.lcoords'))';
%shift
P2.coords=P2.coords+repmat([x y z],size(P2.coords,1),1);P2.lcoords=P2.lcoords+repmat([x y z],size(P2.lcoords,1),1);
val=kernel_3D_exp_prod_sb_emulate(P1,P2,options);%fmeasure(P1,P2,options);
end


%special version operated with global variables
function [val,P2]= pocket_dist_rot_glob(x)
phi=x(1);psi=x(2);xi=x(3);
%rotation matrix
Rf=eye(3,3);R=zeros(3,3);
R(1:2,1:2)=[cos(phi) sin(phi);-sin(phi) cos(phi)];R(3,3)=1; Qphi=R;Rf=Rf*R;
R(:)=0;R(2:3,2:3)=[cos(psi) sin(psi);-sin(psi) cos(psi)];R(1,1)=1;Qpsi=R;Rf=Rf*R;
R(:)=0;R([1 3],[1 3])=[cos(xi) sin(xi);-sin(xi) cos(xi)];R(2,2)=1;Qxi=R;Rf=Rf*R;

global P2;global P1;global options;global fmeasure;
P2l=P2;

P2l.coords=(Rf*(P2l.coords'))';P2l.lcoords=(Rf*(P2l.lcoords'))';
val=fmeasure(P1,P2l,options);
end
%this function calculate the gradient of the difference between  two pockets under given
%rotation of the second pocket: input just two pockets
function grad_rot= pocket_dist_rot_shift_grad(P1,P2,fmeasure,phi,psi,xi,x,y,z,options)
M=size(P2.coords,1);
N=size(P1.coords,1);
%rotation matrix
Rf=eye(3,3);R=zeros(3,3);
R(1:2,1:2)=[cos(phi) sin(phi);-sin(phi) cos(phi)];R(3,3)=1; Qphi=R;Rf=Rf*R;
R(:)=0;R(2:3,2:3)=[cos(psi) sin(psi);-sin(psi) cos(psi)];R(1,1)=1;Qpsi=R;Rf=Rf*R;
R(:)=0;R([1 3],[1 3])=[cos(xi) sin(xi);-sin(xi) cos(xi)];R(2,2)=1;Qxi=R;Rf=Rf*R;

Qphi_prim=zeros(3,3);Qphi_prim(1:2,1:2)=[-sin(phi) cos(phi); -cos(phi) -sin(phi)];
Qpsi_prim=zeros(3,3);Qpsi_prim(2:3,2:3)=[-sin(psi) cos(psi); -cos(psi) -sin(psi)];
Qxi_prim=zeros(3,3);Qxi_prim([1 3],[1 3])=[-sin(xi) cos(xi); -cos(xi) -sin(xi)];

%grad_ang=[Qphi_prim*Qpsi*Qxi+Qphi*Qpsi_prim*Qxi+Qphi*Qpsi*Qxi_prim];
grad_ang=[Qphi_prim*Qpsi*Qxi*P2.coords';Qphi*Qpsi_prim*Qxi*P2.coords';Qphi*Qpsi*Qxi_prim*P2.coords'];%matrix 9*M - derivative of the rotation matrix
ind=repmat((1:M),N,1);ind=ind(:);
%grad_shift([1 4 9],:)=P2.coords';

P2.coords=(Rf*(P2.coords'))';
P2.coords=P2.coords+repmat([x y z],size(P2.coords,1),1);
grad_mes=options.fmeasure_grad(P1,P2,options);% it should be a matrix 3 x N*M 


grad_rot=repmat(grad_mes,3,1).*grad_ang(:,ind);
grad_rot=sum(grad_rot,2);
grad_rot=reshape(grad_rot,3,3);
grad_rot=-sum(grad_rot)'; %minus because P1-RP2

grad_shift=grad_mes;
grad_shift=-sum(grad_shift,2);
%grad_shift=reshape(grad_shift,3,3);
%grad_shift=-sum(grad_shift)'; %minus because P1-RP2
grad_rot=[grad_rot;grad_shift];
end%function

function grad_rot= pocket_dist_rot_grad(P1,P2,fmeasure,phi,psi,xi,options)
M=size(P2.coords,1);
N=size(P1.coords,1);
%rotation matrix
Rf=eye(3,3);R=zeros(3,3);
R(1:2,1:2)=[cos(phi) sin(phi);-sin(phi) cos(phi)];R(3,3)=1; Qphi=R;Rf=Rf*R;
R(:)=0;R(2:3,2:3)=[cos(psi) sin(psi);-sin(psi) cos(psi)];R(1,1)=1;Qpsi=R;Rf=Rf*R;
R(:)=0;R([1 3],[1 3])=[cos(xi) sin(xi);-sin(xi) cos(xi)];R(2,2)=1;Qxi=R;Rf=Rf*R;

Qphi_prim=zeros(3,3);Qphi_prim(1:2,1:2)=[-sin(phi) cos(phi); -cos(phi) -sin(phi)];
Qpsi_prim=zeros(3,3);Qpsi_prim(2:3,2:3)=[-sin(psi) cos(psi); -cos(psi) -sin(psi)];
Qxi_prim=zeros(3,3);Qxi_prim([1 3],[1 3])=[-sin(xi) cos(xi); -cos(xi) -sin(xi)];

%grad_ang=[Qphi_prim*Qpsi*Qxi+Qphi*Qpsi_prim*Qxi+Qphi*Qpsi*Qxi_prim];
grad_ang=[Qphi_prim*Qpsi*Qxi*P2.coords';Qphi*Qpsi_prim*Qxi*P2.coords';Qphi*Qpsi*Qxi_prim*P2.coords'];%matrix 9*M - derivative of the rotation matrix
ind=repmat((1:M),N,1);ind=ind(:);
grad_shift([1 4 9],:)=P2.coords';

P2.coords=(Rf*(P2.coords'))';
grad_mes=options.fmeasure_grad(P1,P2,options);% it should be a matrix 3 x N*M 



grad_rot=repmat(grad_mes,3,1).*grad_ang(:,ind);
grad_rot=sum(grad_rot,2);
grad_rot=reshape(grad_rot,3,3);
grad_rot=-sum(grad_rot)'; %minus because P1-RP2
end%function


%Hessian matrix for objective function
function H= pocket_dist_rot_hessian(P1,P2,fmeasure,phi,psi,xi,options)
M=size(P2.coords,1);
N=size(P1.coords,1);
%rotation matrix
Rf=eye(3,3);R=zeros(3,3);
R(1:2,1:2)=[cos(phi) sin(phi);-sin(phi) cos(phi)];R(3,3)=1; Qphi=R;Rf=Rf*R;
R(:)=0;R(2:3,2:3)=[cos(psi) sin(psi);-sin(psi) cos(psi)];R(1,1)=1;Qpsi=R;Rf=Rf*R;
R(:)=0;R([1 3],[1 3])=[cos(xi) sin(xi);-sin(xi) cos(xi)];R(2,2)=1;Qxi=R;Rf=Rf*R;

Qphi_prim=zeros(3,3);Qphi_prim(1:2,1:2)=[-sin(phi) cos(phi); -cos(phi) -sin(phi)];
Qpsi_prim=zeros(3,3);Qpsi_prim(2:3,2:3)=[-sin(psi) cos(psi); -cos(psi) -sin(psi)];
Qxi_prim=zeros(3,3);Qxi_prim([1 3],[1 3])=[-sin(xi) cos(xi); -cos(xi) -sin(xi)];

Qphi_prim2=-Qphi;Qphi_prim2(3,3)=0;
Qpsi_prim2=-Qpsi;Qpsi_prim2(1,1)=0;
Qxi_prim2=-Qxi;Qxi_prim2(2,2)=0;

%first term (grad((grad_g f))' grad_y g)', grad_y g is a matrix y -colummns,
%g-rows
%grad_y g matrix reshaped into a column
P2t=P2;
P2.coords=(Rf*(P2.coords'))';

grad_ang=[Qphi_prim*Qpsi*Qxi*P2t.coords';Qphi*Qpsi_prim*Qxi*P2t.coords';Qphi*Qpsi*Qxi_prim*P2t.coords'];%matrix 9*M - first derivative of the rotation matrix
%now second derivative of function f
grad_mes=options.fmeasure_grad(P1,P2,options);
ind=repmat((1:M),N,1); ind=ind(:);
grad2f=(repmat(P1.coords,M,1)-P2.coords(ind,:));
grad_ang=grad_ang(:,ind);
ind_rep=1:3;ind_rep=repmat(ind_rep,3,1);ind_rep=ind_rep(:);
grad2f=repmat(grad2f,1,3).*grad2f(:,ind_rep);
grad2f=grad2f'./options.exp_sigma^2;
grad2f([1 4 9],:)=grad2f([1 4 9],:)-1;
grad2f=-repmat(grad_mes,3,1).*grad2f;
%multiplie grad f par grad g the first time, sans transposition
h1=zeros(size(grad2f));
for i=1:3
    for j=1:3
        h1(3*(1-1)+i,:)=sum(grad2f(i:3:end,:).*grad_ang((j-1)*3+1:j*3,:));
    end
end
grad2f=h1;
%transpose f
%ind_rep=reshape(1:9,3,3)'; ind_rep=ind_rep(:);

%now we have multiple matrices presented in columns, also without
%transposition
h1=zeros(size(grad2f));
for i=1:3
    for j=1:3
        h1(3*(1-1)+i,:)=sum(grad2f(i:3:end,:).*grad_ang((j-1)*3+1:j*3,:));
    end
end

%second term
grad_ang=[Qphi_prim2*Qpsi*Qxi*P2t.coords';Qphi_prim*Qpsi_prim*Qxi*P2t.coords';Qphi_prim*Qpsi*Qxi_prim*P2t.coords';...
            Qphi_prim*Qpsi_prim*Qxi*P2t.coords';Qphi*Qpsi_prim2*Qxi*P2t.coords';Qphi*Qpsi_prim*Qxi_prim*P2t.coords';...
            Qphi_prim*Qpsi*Qxi_prim*P2t.coords';Qphi*Qpsi_prim*Qxi_prim*P2t.coords';Qphi*Qpsi*Qxi_prim2*P2t.coords'];%matrix 27*M - derivative of the rotation matrix
%we should transpose it and then take a product with grad_f, but because we
%are working with column represntation, we do not transpose it
grad_mes=options.fmeasure_grad(P1,P2,options).*(repmat(P1.coords,M,1)-P2.coords(ind,:))';
grad_ang=grad_ang(:,ind);
h2=zeros(size(h1));
for i=1:9
   h2(i,:)= sum(grad_mes.*grad_ang((i-1)*3+1:i*3,:));
end
H=h1+h2;
H=sum(H,2);H=reshape(H,3,3);

end%function hessien

% ONE POSSIBLE MEASURE FUNCTION
function val=esp_field_diff(P1,P2,options)
%integration area
x_max=max([P2.coords;P1.coords]);
x_min=min([P2.coords;P1.coords]);
%integral estimation
val=triplequad(@esp_diff,x_min(1),x_max(1),x_min(2),x_max(2),x_min(3),x_max(3),options.tol);

%objective function 
function val=esp_diff(x,y,z)
e0=8.854e-12*1e+19*10^-10;
if (size(x,2)>1) x=x';end;
N1=size(P1.coords,1);N2=size(P2.coords,1);
y=repmat(y,length(x),1); z=repmat(z,length(x),1);
x=[x y z];
%first field
ind_p=repmat((1:N1),size(x,1),1);
r=sqrt(sum((repmat(x,N1,1)-P1.coords(ind_p(:),:))'.^2))';
r=reshape(r,size(x,1),N1);
val=sum((repmat(P1.labels_num',size(x,1),1)./r)');
%second field
ind_p=repmat((1:N2),size(x,1),1);
r=sqrt(sum((repmat(x,N2,1)-P2.coords(ind_p(:),:))'.^2))';
r=reshape(r,size(x,1),N2);

val=abs(val-sum((repmat(P2.labels_num',size(x,1),1)./r)'))/(4*pi*e0);
end

end


%%%% Kernel based approach for MEASURE FUNCTION, rbf kernel for distances, lin kernel for charges
function val=kernel_3D_exp_prod(P1,P2,options)
N=size(P1.coords,1);M=size(P2.coords,1);
ind=repmat(1:M,N,1);ind=ind(:);
K3D=reshape(exp(-sum((repmat(P1.coords,M,1)-P2.coords(ind,:)).^2,2)/(2*options.exp_sigma^2))',N,M).*options.Kp(P1.labels_num,P2.labels_num).*(P1.weights'*P2.weights);
val=-sum(K3D(:));
end
%%%% sitesbase emulation
function val=kernel_3D_exp_prod_sb_emulate(P1,P2,options)
N=size(P1.coords,1);M=size(P2.coords,1);
ind=repmat(1:M,N,1);ind=ind(:);
D=reshape(sum((repmat(P1.coords,M,1)-P2.coords(ind,:)).^2,2)',N,M)<1;
H=hungarian_c(1-D);
val=-trace(D'*H);
end


%%%% Kernel based approach for MEASURE FUNCTION, laplacian kernel for distances, lin kernel for charges
function val=kernel_3D_lap_prod(P1,P2,options)
N=size(P1.coords,1);M=size(P2.coords,1);
ind=repmat(1:M,N,1);ind=ind(:);
K3D=reshape(exp(-sum(abs(repmat(P1.coords,M,1)-P2.coords(ind,:)),2)/(2*options.exp_sigma^2))',N,M).*options.Kp(P1.labels_num,P2.labels_num).*(P1.weights'*P2.weights);
val=-sum(K3D(:));
end

%hard max kernel (not a definite-positive)
function val=kernel_3D_exp_prod_max(P1,P2,options)
N=size(P1.coords,1);M=size(P2.coords,1);
ind=repmat(1:M,N,1);ind=ind(:);
K3D=reshape(exp(-sum((repmat(P1.coords,M,1)-P2.coords(ind,:)).^2,2)/(2*options.exp_sigma^2))',N,M).*options.Kp(P1.labels_num,P2.labels_num).*(P1.weights'*P2.weights);
%val=-sum(K3D(:));
val=-sum(max(K3D'));
end

%%%% generated 3D distance
function val=distance_3D_exp_prod(P1,P2,options)
val=options.add_factor+2*kernel_3D_exp_prod(P1,P2,options);
end

function val=distance_norm_3D_exp_prod(P1,P2,options)
val=2+2*kernel_3D_exp_prod(P1,P2,options)/options.norm_factor;
end
%%%% Kernel gradient: -F(x)x/(sigma^2)
function val=kernel_3D_exp_prod_grad(P1,P2,options)
N=size(P1.coords,1);M=size(P2.coords,1);
ind=repmat(1:M,N,1);ind=ind(:);
K3D=repmat(exp(-sum((repmat(P1.coords,M,1)-P2.coords(ind,:)).^2,2)/(2*options.exp_sigma^2)).*reshape(options.Kp(P1.labels_num,P2.labels_num).*(P1.weights'*P2.weights),length(P1.labels_num)...
    *length(P2.labels_num),1),1,3);
K3D=K3D.*(repmat(P1.coords,M,1)-P2.coords(ind,:));
val=K3D'./(options.exp_sigma^2);
end

%%%% Kernel gradient: -F(x)|x|/(2*sigma^2)
function val=kernel_3D_lap_prod_grad(P1,P2,options)
N=size(P1.coords,1);M=size(P2.coords,1);
ind=repmat(1:M,N,1);ind=ind(:);
K3D=repmat(exp(-sum(abs(repmat(P1.coords,M,1)-P2.coords(ind,:)),2)/(2*options.exp_sigma^2)).*reshape(options.Kp(P1.labels_num,P2.labels_num).*(P1.weights'*P2.weights),length(P1.labels_num)...
    *length(P2.labels_num),1),1,3);
K3D=K3D.*sign(repmat(P1.coords,M,1)-P2.coords(ind,:));
val=K3D'./(2*options.exp_sigma^2);
end

%Hessien F''R'xR'x+F'R''x
function val=kernel_3D_exp_prod_hessien(P1,P2,options)

end

function val=distance_3D_exp_prod_grad(P1,P2,options)
val=2*kernel_3D_exp_prod_grad(P1,P2,options);
end

function val=distance_norm_3D_exp_prod_grad(P1,P2,options)
val=2*kernel_3D_exp_prod_grad(P1,P2,options)/options.norm_factor;
end


